from datetime import datetime, timedelta
from streaming_data_types import serialise_pl72
import platform
import os
from zlib import adler32
from random import randint


def generate_job_id() -> str:
    timestamp = int(datetime.now().timestamp())
    partial_id = "{:s}-{:d}-{:08X}-{:04X}-".format(
        platform.node(), os.getpid(), timestamp ^ 0xFFFFFFFF, randint(0, 65535)
    )
    return partial_id + "{:08X}".format(adler32(partial_id.encode()))


class WriteJob:
    """
    Represents a file-writer write job (before it has been started).
    """

    def __init__(
        self,
        nexus_structure: str,
        file_name: str,
        broker: str,
        start_time: datetime,
        stop_time: datetime = None,
        instrument_name: str = "",
        run_name: str = "",
        metadata: str = "",
    ):
        self.structure = nexus_structure
        self.file = file_name
        self.job_id = generate_job_id()
        self.start = start_time
        if stop_time is None:
            self.stop = self.start + timedelta(days=365.25 * 10)
        else:
            self.stop = stop_time
        self._service_id = ""
        self.broker = broker
        self.instrument_name = instrument_name
        self.run_name = run_name
        self.metadata = metadata

    def generate_new_job_id(self):
        """
        Generate a new job id. Should be called if an attempt at starting this write job fails and another attempt is made.
        """
        self.job_id = generate_job_id()

    @property
    def service_id(self) -> str:
        """
        The service identifier that should process this job. Defaults to an empty string.
        .. note:: Must be set if job is to be processed by a specific file-writer instance.
        """
        return self._service_id

    @service_id.setter
    def service_id(self, new_service_id: str):
        self._service_id = new_service_id

    def get_start_message(self) -> bytes:
        """
        Generate the (flatbuffer) start message that will start this job. If you are sending the message to a specific
        file-writer instance, you have to set the service_id property first or the message will fail to start the job.
        :return: A flatbuffer message that holds the necessary information for starting a write job.
        """
        return serialise_pl72(
            self.job_id,
            self.file,
            self.start,
            self.stop,
            nexus_structure=self.structure,
            service_id=self.service_id,
            broker=self.broker,
            instrument_name=self.instrument_name,
            run_name=self.run_name,
            metadata=self.metadata,
        )



from streaming_data_types.action_response_answ import (
    ActionResponse,
    ActionOutcome,
    ActionType,
)
from json import loads
from file_writer_control.CommandStatus import CommandState
from file_writer_control.JobStatus import JobState
from streaming_data_types.status_x5f2 import StatusMessage
from file_writer_control.WorkerStatus import WorkerState
from typing import Optional


def extract_worker_state_from_status(status: StatusMessage) -> WorkerState:
    """
    Determine the worker state (i.e. file-writer state) based on a file-writer status message.
    :param status: A status update message from a file-writer.
    :return: The extracted worker state.
    """
    json_struct = loads(status.status_json)
    status_map = {"writing": WorkerState.WRITING, "idle": WorkerState.IDLE}
    try:
        status_string = json_struct["state"]
        return status_map[status_string]
    except KeyError:
        return WorkerState.UNKNOWN


def extract_state_from_command_answer(answer: ActionResponse) -> CommandState:
    """
    Determine the command state from a action response message.
    :param answer: The action (either "start a job" or "set top time") response from a file-writer.
    :return: The extracted command state/response.
    """
    status_map = {
        ActionOutcome.Failure: CommandState.ERROR,
        ActionOutcome.Success: CommandState.SUCCESS,
    }
    try:
        return status_map[answer.outcome]
    except KeyError:
        return CommandState.ERROR


def extract_job_state_from_answer(answer: ActionResponse) -> Optional[JobState]:
    """
    Determine the file writing job state from a action response message.
    :param answer: The action (either "start a job" or "set top time") response from a file-writer.
    :return: The extracted job state, None if job state can not be determined from this answer.
    """
    if answer.action == ActionType.StartJob:
        if answer.outcome == ActionOutcome.Success:
            return JobState.WRITING
        else:
            return JobState.ERROR
    return None



from zlib import adler32
from datetime import datetime
import os
import platform
from random import randint


def generate_command_id(command_name: str) -> str:
    """
    Generate a (unique) command identifier.
    :param command_name: The "name" of the command. Will be encoded as part of the command identifier.
    :return: The generated command identifier.
    """
    timestamp = int(datetime.now().timestamp())
    partial_id = f"{platform.node():s}-{os.getpid():d}-{command_name:s}-{timestamp ^ 0xFFFFFFFF:08X}-{randint(0, 65535):04X}-"
    return partial_id + f"{adler32(partial_id.encode()):08X}"



from file_writer_control.WorkerFinder import WorkerFinder
from file_writer_control.WriteJob import WriteJob
from file_writer_control.CommandHandler import CommandHandler
from datetime import datetime
from file_writer_control.JobStatus import JobState
from typing import Union


class JobHandler:
    """
    A stand in for controlling and checking the state of a job running on a file-writer instance.
    """

    def __init__(self, worker_finder: WorkerFinder, job_id=""):
        """
        Constructor.
        :param worker_finder: An instance of a class that inherits from WorkerFinder and implements the member function
        try_start_job of that class.
        :param job_id: (Optional) The job identifier of an existing job.
        """
        self.worker_finder = worker_finder
        self._job_id = job_id

    def start_job(self, job: WriteJob) -> CommandHandler:
        """
        Start a write job. This call is not blocking. It does not guarantee that the write job will actually be started.
        :param job: The write to be started.
        .. note:: Starting a new job will cause the current instance of this class to no longer being able to track or
        control previous jobs.
        :return: A CommandHandler instance that can be used to monitor the outcome of the attempt to start a write job.
        """
        self._job_id = job.job_id
        return self.worker_finder.try_start_job(job)

    def get_state(self) -> JobState:
        """
        Get the state of the job.
        """
        return self.worker_finder.get_job_state(self._job_id)

    def is_done(self) -> bool:
        """
        :return: True if job was completed without errors. False otherwise.
        .. note:: If the job was completed with errors, this call will return False.
        """
        return self.worker_finder.get_job_state(self._job_id) == JobState.DONE

    def get_message(self) -> str:
        """
        Get a string describing the error that was encountered when running the job. (If there was an error.)
        """
        current_status = self.worker_finder.get_job_status(self._job_id)
        if current_status is None:
            return ""
        return current_status.message

    def set_stop_time(self, stop_time: datetime) -> Union[CommandHandler, None]:
        """
        Set a new stop time for the file-writing job. There is no guarantee that the stop time will actually be changed.
        This call is not blocking. Calling this member function will have no effect on the stop-time before the write
        job has started.

        :param stop_time: The new stop time of the job.
        :return: A CommandHandler instance that can be used to monitor the outcome of the attempt to set a new stop time.
        """
        current_status = self.worker_finder.get_job_status(self._job_id)
        if current_status is None:
            return None
        return self.worker_finder.try_send_stop_time(
            current_status.service_id, self._job_id, stop_time
        )

    def stop_now(self) -> Union[CommandHandler, None]:
        """
        Tell the file-writing to stop writing NOW. There is no guarantee that will actually happen though.
        This call is not blocking. Calling this member function will have no effect if done before a write job has
        actually started.

        :return: A CommandHandler instance that can be used to monitor the outcome of the attempt to set a new stop time.
        """
        current_status = self.worker_finder.get_job_status(self._job_id)
        if current_status is None:
            return None
        return self.worker_finder.try_send_stop_now(
            current_status.service_id, self._job_id
        )



from enum import Enum, auto
from datetime import datetime


class CommandState(Enum):
    """
    The state of a command.
    """

    UNKNOWN = auto()
    NO_COMMAND = auto()
    WAITING_RESPONSE = auto()
    TIMEOUT_RESPONSE = auto()
    ERROR = auto()
    SUCCESS = auto()


class CommandStatus(object):
    """
    The status of a command.
    """

    def __init__(self, job_id: str, command_id: str):
        self._job_id = job_id
        self._command_id = command_id
        self._last_update = datetime.now()
        self._state = CommandState.NO_COMMAND
        self._message = ""

    def __eq__(self, other_status: "CommandStatus"):
        if not isinstance(other_status, CommandStatus):
            raise NotImplementedError
        return (
            other_status.command_id == self.command_id
            and other_status.job_id == self.job_id
            and other_status.state == self.state
        )

    def update_status(self, new_status: "CommandStatus"):
        """
        Updates the status/state of this instance of the CommandStatus class using another instance.
        .. note:: The command identifier of both this instance and the other one must be identical.
        :param new_status: The other instance of the CommandStatus class.
        """
        if new_status.command_id != self.command_id:
            raise RuntimeError(
                f"Command id of status update is not correct ({self.command_id} vs {new_status.command_id})"
            )
        self._state = new_status.state
        if new_status.message:
            self._message = new_status.message
        self._last_update = new_status.last_update

    @property
    def job_id(self) -> str:
        """
        The job identifier under which this command is executed.
        """
        return self._job_id

    @property
    def command_id(self) -> str:
        """
        The unique command identifier of this command.
        """
        return self._command_id

    @property
    def message(self) -> str:
        """
        A status/error message as returned by the file-writer that responded to the command.
        :return:
        """
        return self._message

    @message.setter
    def message(self, new_message: str):
        if new_message:
            self._message = new_message
            self._last_update = datetime.now()

    @property
    def state(self) -> CommandState:
        """
        The current state of the command.
        """
        return self._state

    @state.setter
    def state(self, new_state: CommandState):
        self._state = new_state
        self._last_update = datetime.now()

    @property
    def last_update(self) -> datetime:
        """
        The local time stamp of the last update of the status of the command.
        """
        return self._last_update



from contextlib import contextmanager
import time


@contextmanager
def sleep_in_between(sleep_time: float):
    """
    This manages the sleep times after! task execution.
    """
    try:
        yield  # Execute the body. Any exception will be catched here.
    finally:
        time.sleep(sleep_time)  # Sleep after execution.



import os
from setuptools import setup, find_packages

DESCRIPTION = (
    "A small Python library for controlling the ESS kafka-to-nexus file-writer."
)

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
try:
    with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except Exception as error:
    print("COULD NOT GET LONG DESC: {}".format(error))
    LONG_DESCRIPTION = DESCRIPTION

setup(
    name="file_writer_control",
    version="1.0.1",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="ScreamingUdder",
    url="https://github.com/ess-dmsc/file_writer_control",
    license="BSD 2-Clause License",
    packages=find_packages(exclude=["tests", "examples"]),
    python_requires=">=3.6.0",
    install_requires=["kafka-python>=2.0", "ess-streaming-data-types>=0.9.5"],
)
